package org.jbjf.tasks;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import org.jbjf.core.AbstractTask;
import org.jbjf.core.file.DelimitedDocumentAdapter;
import org.jbjf.core.file.FlatRow;
import org.jbjf.plugin.IJBJFPluginCipher;
import org.jbjf.plugin.IJBJFPluginDatabase;
import org.jbjf.xml.JBJFDatabaseConnection;
import org.jbjf.xml.JBJFPluginDefinition;
import org.jbjf.xml.JBJFSQLDefinition;
import org.jbjf.xml.JBJFSQLParameter;

/**
 * <p>
 * The <code>LoadDelimitedDocument</code> Task is designed to pull
 * a <code>DelimitedDocumentAdapter</code> object from the Job 
 * Stack and load that text file into an SQL database.  See the Input
 * Resources for details on the &lt;resource&gt; elements needed.
 * </p>
 * <p>
 * <h3>Details</h3>
 * <hr>
 * <h4>Input Resources</h4>
 * <table border='1' width='100%'>
 * <thead>
 *  <tr>
 *      <td width='15%'>Resource</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td width='7%'>Required</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td>Type</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td>Description/Comments</td>
 *  </tr>
 * </thead>
 *  <tr valign='top'>
 *      <td>Directory Path</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A partial or full directory path of where the file resides.  When
 *      not provided, defaults to . (current working directory).
 *      </td>
 *  </tr>
 *  <tr valign='top'>
 *      <td>Delimiter</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A single character that indicates a text delimiter in the
 *      text file between the columns of data.  When not provided,
 *      defaults to , (comma).
 *      </td>
 *  </tr>
 * </table>
 * </p>
 * <p>
 * <pre>
 * </pre>
 * </p>
 * <p>
 * @author  Adym S. Lincoln<br>
 * Copyright (C) 2007-2013. JBJF All rights reserved.
 * @version 2.0.0
 * @since   1.0.0
 * </p>
 * <p>
 * @see org.jbjf.core.AbstractTask
 * </p>
 */

public class LoadDelimitedDocument extends AbstractTask {
    /** 
     * Stores a fully qualified class name.  Used for debugging and 
     * auditing.
     * @since 1.0.0
     */
    public static final String ID = LoadDelimitedDocument.class.getName();

    /** 
     * Stores the class name, primarily used for debugging and so 
     * forth.  Used for debugging and auditing.
     * @since 1.0.0
     */
    private String SHORT_NAME = "LoadDelimitedDocument()";

    /** 
     * Stores a <code>SYSTEM IDENTITY HASHCODE</code>.  Used for
     * debugging and auditing.
     * @since 1.0.0
     */
    private String SYSTEM_IDENTITY = String.valueOf(System.identityHashCode(this));

    /**
     * Class property that stores the task's SQL Definition
     * object.
     */
    protected   JBJFSQLDefinition           sqlDefinition;

    /** 
     * Class property that stores a connection definition resource.
     */
    protected   JBJFDatabaseConnection      connectionDefinition;

    /**
     * Class property that stores the type attribute for the specific
     * plugin-database to be used for lookup in the task plugin cache.
     */
    protected   JBJFPluginDefinition        pluginDatabaseDefinition;
    
    /**
     * Class property that stores the type attribute for the specific
     * plugin-cipher to be used for lookup in the task plugin cache.
     */
    protected   JBJFPluginDefinition        pluginCipherDefinition;

    /**
     * Class property that stores the Database Plugin runtime object.  This 
     * is stored at the class level to allow the <code>runTask()</code>
     * method the ability to close the database plugin's 
     * <code>java.sql.Connection</code> object using <code>closeConnection()</code>.
     */
    protected   IJBJFPluginDatabase         pluginDatabaseRuntime = null;

    /**
     * Class property that stores the Cipher Plugin runtime object.  This
     * may be null if there is no Cipher being used.
     */
    protected   IJBJFPluginCipher           pluginCipherRuntime = null;

    /**
     * Class property that stores the Job Stack key where the 
     * <code>DelimitedDocumentAdapter</code> is stored.
     */
    protected   String                      jobStackKey;
    
    /**
     * Class property that stores the <code>DelimitedDocumentAdapter</code>
     * that contains the text file data we want to load.
     */
    protected   DelimitedDocumentAdapter    delimitedDocument;

    /**
     * Class property that stores the total row count processed.
     */
    protected   long                        totalRowCount = 0;

    /**
     * Class property that stores the row count of failed transactions.
     */
    protected   long                        failureRowCount = 0;

    /**
     * Class property that stores the row count of successful transactions.
     */
    protected   long                        successRowCount = 0;

    /**
     * Class property that manages the JDBC connection object.
     */
    protected   Connection                  connection;

    private String REZ_KEY_SQLDEFINITION    = "sql-definition";
    private String REZ_KEY_CONNECTION       = "connection";
    private String REZ_KEY_PLUGINDATABASE   = "plugin-database";
    private String REZ_KEY_PLUGINCIPHER     = "plugin-cipher";
    private String REZ_KEY_DELIMITEDDOC     = "delimited-document";
    private String REZ_KEY_JOBSTACK         = "job-stack-key";

    /**
     * Default constructor.  Required for JBJF integration.
     */
    public LoadDelimitedDocument() {
        // CHECK : Abstract classes need to always check to make sure
        // a parent class has NOT already initialized the required
        // resources.
        if ( getRequiredResources() == null ) { 
            mtaskRequired = new ArrayList();
        }
        getRequiredResources().add("sql-definition");   /* sql statement to run */
        getRequiredResources().add("connection");       /* database connection  */
        getRequiredResources().add("plugin-database");  /* JDBC management      */
    }

    /* (non-Javadoc)
     * @see org.jbjf.core.AbstractTask#runTask(java.util.HashMap)
     */
    @Override
    public void runTask ( HashMap jobStack ) throws Exception {
        getLog().info( SHORT_NAME + "...Start..." );
        
        /* Re-push the job stack onto the job stack to ensure that
         * no additional parameters were added since initTask() ran.
         */
        this.mjobStack = jobStack;

        if ( hasRequiredResources() ) {
            try { 
                /* RESOLVE RESOURCES : Locate and store the individual
                 * objects from the Job Stack into the class properties...
                 */
                setDelimitedDocument ( 
                    (DelimitedDocumentAdapter) getParameters ().get ( 
                        (String)getResources().get ( REZ_KEY_DELIMITEDDOC )
                    )
                );
                setJobStackKey ( (String)getResources().get ( REZ_KEY_JOBSTACK ) );
                
                /*
                 * Use the job stack key to fetch the database
                 * connection...when using TEMPDB tables you'll need
                 * to maintain the connection/session across multiple tasks.
                 */
                if ( getConnection() == null ) {
                    throw new Exception ( 
                        "A database connection could not be established.  Please"
                        + " check all properties in the <connection> element"
                        + " named [" + getConnectionDefinition().getName () + "]"
                    );
                }
//                else if ( getConnection().isClosed() ) {
//                    throw new Exception ( 
//                        "The database connection is closed"
//                        + "...Please check the"
//                        + " Batch Definition file and make sure the"
//                        + " connection was opened successfully in a"
//                        + " previous task."
//                    );
//                }
                else {
                    Statement   ldtbStatement = null;
                    // STATEMENT : Create the database statement...
                    //
                    ldtbStatement = getConnection().createStatement ();
                    
                    // LEADING ZERO : Provides a simple decimal format
                    // with leading seros that allows us to iterate
                    // thru all the JBJFSQLParamaters, p001, p002, p003, etc...
                    // 
                    DecimalFormat   leadingZeros = new DecimalFormat ( "000" );

                    int             rowCount = 0;
                    int             successCount = 0;
                    
                    // ROWS : Iterate thru all the rows in the document...
                    // 
                    for ( int lngIdx=0; lngIdx < getDelimitedDocument().getRecordCount(); lngIdx++ ) {
                        FlatRow vmRow = (FlatRow)getDelimitedDocument().getRecordContainer(lngIdx);
                        // COLUMNS : Iterate thru all the columns in the row...
                        // 
                        for ( int colIdx=0; colIdx<vmRow.getColumnCount(); colIdx++ ) {
                            // sql-definition parms are "1" based...
                            // build a lookup key to the current SQL
                            // parameter...p00n...
                            // 
                            String lparmKey = "p" + leadingZeros.format(colIdx+1);
                            // Use the lookup key to find the parameter object...
                            // 
                            JBJFSQLParameter    ltheParm = (JBJFSQLParameter)getSqlDefinition().getParameters().get(lparmKey);
                            // Grab the value from the document row column...
                            String              ltheValue = vmRow.getColumnValue(colIdx);
                            if ( ltheParm != null ) { 
                                if ( ltheValue != null ) { ltheValue = scrubData ( ltheValue ); }
                                // Set the value of the parameter using the
                                // column value...
                                ltheParm.setValue (
                                    ltheValue
                                );
                            }
                        }

                        // SQL : Resolve parameters and fetch the SQL...
                        // This will take all the SQL Parameter values 
                        // and substitute them into the SQL Statement...
                        //
                        String lstrSQL = getSqlDefinition().getStatement();

                        // NULL : Force all 'null' to true NULL...
                        // 
                        lstrSQL = lstrSQL.replaceAll( "'null'", "null" ); 

                        // RUN : Run the SQL against the database...
                        //
                        getLog().debug( SHORT_NAME + " SQL " );
                        getLog().debug( "-------------------------------------------------" );
                        getLog().debug( lstrSQL );
                        getLog().debug( "-------------------------------------------------" );
                        try {
                            ldtbStatement.executeUpdate ( lstrSQL );
                            successCount++;
                        }
                        catch ( SQLException lsqlXcp ) {
                            getLog().warn ( 
                                SHORT_NAME 
                                + " SQLException on Record [" 
                                + rowCount + "]" 
                            );
                            getLog().warn( lstrSQL );
                        }
                        catch ( Exception ltheXcp ) {
                            throw ltheXcp;
                        }
                        rowCount++;
                    }

                    // CHECK : If the rowcount doesn't match the
                    // success count, then some records didn't load...
                    //
                    if ( rowCount != successCount ) {
                        getLog().warn ( 
                            SHORT_NAME 
                            + " Not all rows were successfully loaded.  " 
                            + successCount + " of " 
                            + rowCount + " were loaded!" 
                        );
                    }
                    else {
                        getLog().info ( 
                            SHORT_NAME 
                            + " Rows were successfully loaded [" 
                            + successCount + "]" 
                        );
                    }

                    // GC : Check for the DelimitedDocument on the job
                    // stack and remove it for GC...
                    // 
                    String  lkeyDoc = (String)getResources().get( REZ_KEY_JOBSTACK );
                    if ( getParameters().containsKey(lkeyDoc) ) {
                        getParameters().remove(lkeyDoc);
                    }
                }

            }
            catch ( Exception ltheXcp ) {
                getLog().fatal(ltheXcp.getMessage());
                throw ltheXcp;
            }
            finally { 
                getPluginDatabaseRuntime().closeConnection ();
            }
        }
        getLog().info( SHORT_NAME + "...Complete..." );

    }

    /* (non-Javadoc)
     * @see org.jbjf.core.AbstractTask#initTask(java.util.HashMap)
     */
    @Override
    public void initTask(HashMap jobStack) throws Exception {
        super.initTask ( jobStack );
        getLog().debug( "Initialize SQL Work Unit...Start..." );

        if ( getResources() != null ) {
            getLog().debug( "Gathering SQL resources..." );
            Iterator    lstepThru   = getResources().entrySet().iterator();
            while ( lstepThru.hasNext() ) {
                Entry ltheItem = (Entry)lstepThru.next();
                String lstrTypeResource = (String)ltheItem.getKey();
                if ( lstrTypeResource.equalsIgnoreCase("sql-definition") ) {
                    getLog().debug( "Gathering " + lstrTypeResource + " [ " + (String)ltheItem.getValue() + " ]" );
                    setSqlDefinition ( 
                        (JBJFSQLDefinition)getDefinition().getSQLDefinitions().get ( 
                            (String)ltheItem.getValue() 
                        )
                    );
                } 
                else if (lstrTypeResource.equalsIgnoreCase("connection")) {
                    setConnectionDefinition (
                        (JBJFDatabaseConnection) getDefinition().getConnections().get(
                            (String) ltheItem.getValue()
                        )
                    );
                } 
            }
            lstepThru = null;
        }

        // PLUGIN-DEFINITION : Database
        // Grab the database plugin definition, this is the 
        // plugin-definition (XML) object, not the actual runtime 
        // object
        // 
        setPluginDatabaseDefinition ( 
            (JBJFPluginDefinition)getResources().get ( "plugin-database" )
        );

        // PLUGIN-DEFINITION : Cipher
        // Grab the OPTIONAL cipher plugin definition, this is the plugin-definition
        // (XML) object, not the actual runtime object
        // 
        if ( getResources().containsKey( "plugin-cipher" ) ) {
            setPluginCipherDefinition ( 
                (JBJFPluginDefinition)getResources().get ( "plugin-cipher" )
            );
        }

        // RUNTIME PLUGINS : Database 
        // Grab the runtime plugins from the task cache...Using
        // the id from the plugin-definition, we can lookup the actual 
        // runtime instances of the plugins.
        // 
        setPluginDatabaseRuntime ( 
            getTaskPlugins().getDatabasePlugin( 
                getPluginDatabaseDefinition().getPluginId() 
            ) 
        );

        // RUNTIME PLUGINS : 
        // Populate the database plugin with the connection
        // definition property values.  The connection definition "defines"
        // the database connection and the database plugin
        // "implements" the JDBC connection.  The JDBC connection 
        // is NOT established yet, only the properties are supplied...
        // 
        if ( getConnectionDefinition() != null ) { 
            getPluginDatabaseRuntime().setProperties( getConnectionDefinition() );
        }
        else { 
            throw new Exception ( 
                SHORT_NAME 
                + " Plugin Database Definition [" + getPluginDatabaseDefinition().getPluginId ()
                + " has no Connection Definition." 
                + "  Please check the resources for the task [" + this.toString () + "]" 
            );
        }

        // RUNTIME PLUGINS : Cipher 
        // Grab the runtime plugins from the task cache...Using
        // the id from the plugin-definition, we can lookup the actual 
        // runtime instances of the plugins.
        // 
        if ( getPluginCipherDefinition() != null ) {
            setPluginCipherRuntime ( 
                getTaskPlugins().getCipherPlugin( 
                    getPluginCipherDefinition().getPluginId() 
                ) 
            );
        }

        getLog().debug( "Initialize SQL Work Unit...Complete..." );
    }

    /**
     * <p>
     * The getConnection() is a local method for the task that works
     * with the Database Plugin to create a <code>java.sql.Connection</code>
     * object that can be used by the task to run SQL statements.
     * The getConnection() (or similar method) really just needs to
     * do the following:
     * <ul>
     * <li>Check for the existence of a Database Runtime Plugin.</li>
     * <li>Check the class java.sql.Connection property for a non-null value.</li>
     * <li>Check the Connection to make sure it's not closed.</li>
     * <li>Check the Connection to make sure it's not closed.</li>
     * </ul>
     * This method is reusable for any SQL task that uses a single 
     * database plugin and optional single cipher plugin.  The method 
     * is dependent on there being a plugin-database and plugin-cipher and that
     * they are listed as &lt;resource&gt; elements in the &lt;task&gt;
     * element:
     * <ul>
     * <li>plugin-database</li>
     * <li>plugin-cipher</li>
     * </ul>
     * </p>
     * <p>
     * If ....
     * </p>
     * <p>
     * @return  A <code>java.sql.Connection</code> object for the
     * task.
     * </p>
     */
    public  Connection  getConnection() throws Exception {
        /* PERSIST : Check and see if the Database plugin still 
         * has a database connection AND if that connection is
         * still open.
         */
        if ( getPluginDatabaseRuntime() != null ) {
            if ( this.connection != null ) {
                if ( this.connection .isClosed() ) {
                    this.connection = null;
                    getPluginDatabaseRuntime().closeConnection ();
                }
            }
            else { 
                this.connection = null;
            }
            
            /* CHECK : Check and see if the Database plugin still 
             * had a database connection AND that the connection was
             * still open.  A non-null value means the connection 
             * was still there AND open.
             */
            if ( this.connection == null ) { 
                // CONNECT : Connect to the database...pass in the cipher
                // plugin to decrypt the userid and password.
                //
                if ( getPluginCipherRuntime() != null ) {
                    this.connection = getPluginDatabaseRuntime().getConnection(getPluginCipherRuntime());
                }
                else {
                    this.connection = getPluginDatabaseRuntime().getConnection();
                }
                
            }

        }

        return this.connection;

    }

    /**
     * Utility method that cleans out special characters that could
     * cause an SQL statement to fail.
     * <p>
     * 
     * @param someData
     *            A <code>String</code> containing the data to cleanse.
     * @return Returns a <code>String</code> value minus single-quote (')
     * or line feeds (\n)
     */
    public String scrubData(String someData) throws Exception {
        String lstrResults = someData;

        if ( lstrResults.indexOf('\'') > -1 ) {
            lstrResults = lstrResults.replace('\'', ' ');
        }
        if ( lstrResults.indexOf('\n') > -1 ) {
            lstrResults = lstrResults.replace('\n', ' ');
        }
        if ( lstrResults.indexOf(',') > -1 ) {
            lstrResults = lstrResults.replace(",","");
        }
        return lstrResults.trim();
    }

    
    public JBJFSQLDefinition getSqlDefinition() {
        return sqlDefinition;
    }

    public void setSqlDefinition(JBJFSQLDefinition sqlDefinition) {
        this.sqlDefinition = sqlDefinition;
    }

    public JBJFDatabaseConnection getConnectionDefinition() {
        return connectionDefinition;
    }

    public void setConnectionDefinition(JBJFDatabaseConnection connectionDefinition) {
        this.connectionDefinition = connectionDefinition;
    }

    public JBJFPluginDefinition getPluginDatabaseDefinition() {
        return pluginDatabaseDefinition;
    }

    public void setPluginDatabaseDefinition(JBJFPluginDefinition pluginDefinition) {
        this.pluginDatabaseDefinition = pluginDefinition;
    }

    public JBJFPluginDefinition getPluginCipherDefinition() {
        return pluginCipherDefinition;
    }

    public void setPluginCipherDefinition(JBJFPluginDefinition pluginDefinition) {
        this.pluginCipherDefinition = pluginDefinition;
    }

    public IJBJFPluginDatabase getPluginDatabaseRuntime() {
        return pluginDatabaseRuntime;
    }

    public void setPluginDatabaseRuntime(IJBJFPluginDatabase pluginDatabase) {
        this.pluginDatabaseRuntime = pluginDatabase;
    }

    public IJBJFPluginCipher getPluginCipherRuntime() {
        return pluginCipherRuntime;
    }

    public void setPluginCipherRuntime(IJBJFPluginCipher pluginCipher) {
        this.pluginCipherRuntime = pluginCipher;
    }

    public String getJobStackKey() {
        return jobStackKey;
    }

    public void setJobStackKey(String jobStackKey) {
        this.jobStackKey = jobStackKey;
    }

    public DelimitedDocumentAdapter getDelimitedDocument() {
        return delimitedDocument;
    }

    public void setDelimitedDocument(DelimitedDocumentAdapter delimitedDocument) {
        this.delimitedDocument = delimitedDocument;
    }

    public long getTotalRowCount() {
        return totalRowCount;
    }

    public void setTotalRowCount(long totalRowCount) {
        this.totalRowCount = totalRowCount;
    }

    public long getFailureRowCount() {
        return failureRowCount;
    }

    public void setFailureRowCount(long failureRowCount) {
        this.failureRowCount = failureRowCount;
    }

    public long getSuccessRowCount() {
        return successRowCount;
    }

    public void setSuccessRowCount(long successRowCount) {
        this.successRowCount = successRowCount;
    }

    public void setConnection(Connection connection) {
        this.connection = connection;
    }

}
