package org.jbjf.tasks;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import org.jbjf.core.AbstractTask;
import org.jbjf.core.file.DelimitedDocument;
import org.jbjf.core.file.DelimitedDocumentAdapter;

/**
 * <p>
 * The <code>ReadTextFile</code> is designed to read a delimited 
 * text file and stores that file as a <code>DelimitedDocumentAdapter</code>
 * object onto the JBJF Job Stack.  Once on the Job Stack, the object
 * can be pulled off and used in other Task classes.
 * </p>
 * <p>
 * <h3>Details</h3>
 * <hr>
 * <p>
 * The <code>ReadTextFile</code> takes the following &lt;resource&gt;
 * elements to operate.  Be attentive to the Required column as some
 * of the resources may not be needed.
 * </p>
 * <h4>Input Resources</h4>
 * <table border='1' width='100%'>
 * <thead>
 *  <tr>
 *      <td width='15%'>Resource</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td width='7%'>Required</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td>Type</td>
 *      <td width='1%'>&nbsp;</td>
 *      <td>Description/Comments</td>
 *  </tr>
 * </thead>
 *  <tr valign='top'>
 *      <td>filename</td>
 *      <td>&nbsp;</td>
 *      <td>Y</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      The name of the text file, no directory path should be 
 *      included.
 *      </td>
 *  </tr>
 *  <tr valign='top'>
 *      <td>path</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A partial or full directory path of where the file resides.  When
 *      not provided, defaults to . (current working directory).
 *      </td>
 *  </tr>
 *  <tr valign='top'>
 *      <td>delimiter</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A single character that indicates a text delimiter in the
 *      text file between the columns of data.  When not provided,
 *      defaults to , (comma).  Special delimiters are coded as 
 *      TAB, etc...
 *      </td>
 *  </tr>
 *  <tr valign='top'>
 *      <td>header</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A true/false indicator on whether there is a header line
 *      that contains the column names.  Defaults to false.
 *      </td>
 *  </tr>
 *  <tr valign='top'>
 *      <td>job.stack.key</td>
 *      <td>&nbsp;</td>
 *      <td>N</td>
 *      <td>&nbsp;</td>
 *      <td>String</td>
 *      <td>&nbsp;</td>
 *      <td>
 *      A unique text string (key) that will be used to store the
 *      outcome/results of the Task onto the Job Stack.  Defaults
 *      to read.text.file.
 *      </td>
 *  </tr>
 * </table>
 * </p>
 * <p>
 * <pre>
 * </pre>
 * </p>
 * <p>
 * @author  Adym S. Lincoln<br>
 * Copyright (C) 2007-2013. JBJF All rights reserved.
 * @version 2.0.0
 * @since   1.0.0
 * </p>
 * <p>
 * @see org.jbjf.core.AbstractTask
 * </p>
 */

public class ReadTextFile extends AbstractTask {

    /** 
     * Stores a fully qualified class name.  Used for debugging and 
     * auditing.
     * @since 1.0.0
     */
    public static final String ID = ReadTextDocument.class.getName();

    /** 
     * Stores the class name, primarily used for debugging and so 
     * forth.  Used for debugging and auditing.
     * @since 1.0.0
     */
    private String SHORT_NAME = "ReadTextDocument()";

    /** 
     * Stores a <code>SYSTEM IDENTITY HASHCODE</code>.  Used for
     * debugging and auditing.
     * @since 1.0.0
     */
    private String SYSTEM_IDENTITY = String.valueOf(System.identityHashCode(this));

    private String REZ_KEY_PATH         = "path";
    private String REZ_KEY_FILE         = "filename";
    private String REZ_KEY_DELIMITER    = "delimiter";
    private String REZ_KEY_JOBSTACK     = "job-stack-key";
    private String REZ_KEY_HEADER       = "header";

    /**
     * Class property that stores the delimiter.  By default, initialized
     * in the initTask() method of the JBJF Task super class.
     */
    protected   String      delimiter = ",";
    
    /**
     * Class property that stores the directory path.  By default, initialized
     * in the initTask() method of the JBJF Task super class.
     */
    protected   String      path = ".";
    
    /**
     * Class property that stores the filename.  By default, initialized
     * in the initTask() method of the JBJF Task super class.
     */
    protected   String      filename = null;
    
    /**
     * Class property that stores the Job Stack key that will be 
     * used to store the return object of the task.  By default, initialized
     * in the initTask() method of the JBJF Task super class.
     */
    protected   String      jobStackKey = "read.text.file";
    
    
    /**
     * Default constructor.  A required constructor for JBJF.
     */
    public ReadTextFile() {
        super ();
        if ( getRequiredResources() == null ) { 
            setRequiredResources ( new ArrayList() );
        }
        getRequiredResources().add(REZ_KEY_FILE);
    }

    /* (non-Javadoc)
     * @see org.jbjf.core.AbstractTask#runTask(java.util.HashMap)
     */
    @Override
    public void runTask(HashMap jobStack) throws Exception {
        /*
         * 
         */
        getLog().info( SHORT_NAME + "...Start..." );
        getLog().info( SHORT_NAME + "...Start...CWD{" + System.getProperty("user.dir") + "]" );
        

        /* CHECK : Make sure the task has all the required resources.
         */
        if ( hasRequiredResources() ) {

            try { 
                String  lkeyHeaders  = null;
                boolean lblnHeader = false;
                if ( getResources().containsKey(REZ_KEY_HEADER) ) {
                    lkeyHeaders = (String)getResources().get( REZ_KEY_HEADER );
                    lblnHeader = lkeyHeaders.contains("true");
                }

                if ( getResources().containsKey(REZ_KEY_PATH) ) {
                    setPath ( (String)getResources ().get ( REZ_KEY_PATH ) );
                }

                if ( getResources().containsKey(REZ_KEY_FILE) ) {
                    setFilename ( (String)getResources ().get ( REZ_KEY_FILE ) );
                }

                if ( getResources().containsKey(REZ_KEY_DELIMITER) ) {
                    setDelimiter ( (String)getResources ().get ( REZ_KEY_DELIMITER ) );
                }

                if ( getResources().containsKey(REZ_KEY_JOBSTACK) ) {
                    setJobStackKey ( (String)getResources ().get ( REZ_KEY_JOBSTACK ) );
                }

                /* RESULTS : Initialize the results object.
                 */
                DelimitedDocumentAdapter    ldlmDocument = null;

                getLog().debug( "Text file name [" + getFilename() + "]" );
                getLog().debug( "Text file source directory [" + getPath() + "]" );

                /* CHAR : The DelimitedDocumentAdapter uses a single
                 * character (char) as the delimiter, we must convert
                 * the String property to a char.
                 */
                char    lchrDelimiter = getDelimiter().charAt( 0 );
                getLog().debug( "Text file delimiter [" + lchrDelimiter + "]" );

                getLog().debug( "Job Stack storage key [" + getJobStackKey() + "]" );

                if ( getResources().containsKey("header") ) {
                    lkeyHeaders = (String)getResources().get( "header" );
                    lblnHeader = lkeyHeaders.contains("true");
                }
                else {
                    lkeyHeaders = "";
                    lblnHeader = false;
                }
                getLog().debug( "Job Stack headers [" + lblnHeader + "]" );

                ldlmDocument = new DelimitedDocumentAdapter ( 
                     getPath() + File.separator + getFilename()
                    ,lchrDelimiter
                    ,lblnHeader 
                );

                getParameters().put(getJobStackKey(), ldlmDocument);

            }
            catch ( Exception ltheXcp ) {
                getLog().fatal(ltheXcp.getMessage());
                throw ltheXcp;
            }
        }

        getLog().info( SHORT_NAME + "...Complete..." );

    }

    public String getDelimiter() {
        return delimiter;
    }

    public void setDelimiter(String delimiter) {
        this.delimiter = delimiter;
    }

    public String getPath() {
        return path;
    }

    public void setPath(String path) {
        this.path = path;
    }

    public String getFilename() {
        return filename;
    }

    public void setFilename(String filename) {
        this.filename = filename;
    }

    public String getJobStackKey() {
        return jobStackKey;
    }

    public void setJobStackKey(String jobStackKey) {
        this.jobStackKey = jobStackKey;
    }



}
